import os
import queue
import torch
import warnings

import numpy as np

from .common import TranslationTask, SAMPLE_RATE, LoopWorkerBase

warnings.filterwarnings('ignore')


def _init_jit_model(model_path: str, device=torch.device('cpu')):
    torch.set_grad_enabled(False)
    model = torch.jit.load(model_path, map_location=device)
    model.eval()
    return model


class VAD:

    def __init__(self):
        current_dir = os.path.dirname(__file__)
        self.model = _init_jit_model(os.path.join(current_dir, 'silero_vad.jit'))

    def is_speech(self, audio: np.array, threshold: float = 0.5, sampling_rate: int = 16000):
        if not torch.is_tensor(audio):
            try:
                audio = torch.Tensor(audio)
            except:
                raise TypeError('Audio cannot be casted to tensor. Cast it manually')
        speech_prob = self.model(audio, sampling_rate).item()
        return speech_prob >= threshold

    def reset_states(self):
        self.model.reset_states()


class AudioSlicer(LoopWorkerBase):

    def __init__(self, frame_duration: float, continuous_no_speech_threshold: float,
                 min_audio_length: float, max_audio_length: float, prefix_retention_length: float,
                 vad_threshold: float):
        self.vad = VAD()
        self.continuous_no_speech_threshold = round(continuous_no_speech_threshold / frame_duration)
        self.min_audio_length = round(min_audio_length / frame_duration)
        self.max_audio_length = round(max_audio_length / frame_duration)
        self.prefix_retention_length = round(prefix_retention_length / frame_duration)
        self.vad_threshold = vad_threshold
        self.sampling_rate = SAMPLE_RATE
        self.audio_buffer = []
        self.prefix_audio_buffer = []
        self.speech_count = 0
        self.no_speech_count = 0
        self.continuous_no_speech_count = 0
        self.frame_duration = frame_duration
        self.counter = 0
        self.last_slice_second = 0.0

    def put(self, audio: np.array):
        self.counter += 1
        if self.vad.is_speech(audio, self.vad_threshold, self.sampling_rate):
            self.audio_buffer.append(audio)
            self.speech_count += 1
            self.continuous_no_speech_count = 0
        else:
            if self.speech_count == 0 and self.no_speech_count == 1:
                self.slice()
            self.audio_buffer.append(audio)
            self.no_speech_count += 1
            self.continuous_no_speech_count += 1
        if self.speech_count and self.no_speech_count / 4 > self.speech_count:
            self.slice()

    def should_slice(self):
        audio_len = len(self.audio_buffer)
        if audio_len < self.min_audio_length:
            return False
        if audio_len > self.max_audio_length:
            return True
        if self.continuous_no_speech_count >= self.continuous_no_speech_threshold:
            return True
        return False

    def slice(self):
        concatenate_buffer = self.prefix_audio_buffer + self.audio_buffer
        concatenate_audio = np.concatenate(concatenate_buffer)
        self.audio_buffer = []
        self.prefix_audio_buffer = concatenate_buffer[-self.prefix_retention_length:]
        self.speech_count = 0
        self.no_speech_count = 0
        self.continuous_no_speech_count = 0
        # self.vad.reset_states()
        slice_second = self.counter * self.frame_duration
        last_slice_second = self.last_slice_second
        self.last_slice_second = slice_second
        return concatenate_audio, (last_slice_second, slice_second)

    def loop(self, input_queue: queue.SimpleQueue[np.array],
             output_queue: queue.SimpleQueue[TranslationTask]):
        while True:
            audio = input_queue.get()
            self.put(audio)
            if self.should_slice():
                sliced_audio, time_range = self.slice()
                task = TranslationTask(sliced_audio, time_range)
                output_queue.put(task)



import os
import queue
from scipy.io.wavfile import write as write_audio

import numpy as np
from openai import OpenAI, DefaultHttpxClient

from . import filters
from .common import TranslationTask, SAMPLE_RATE, LoopWorkerBase, sec2str

TEMP_AUDIO_FILE_NAME = '_whisper_api_temp.wav'


def _filter_text(text: str, whisper_filters: str):
    filter_name_list = whisper_filters.split(',')
    for filter_name in filter_name_list:
        filter = getattr(filters, filter_name)
        if not filter:
            raise Exception('Unknown filter: %s' % filter_name)
        text = filter(text)
    return text


class OpenaiWhisper(LoopWorkerBase):

    def __init__(self, model: str, language: str) -> None:
        print('Loading whisper model: {}'.format(model))
        import whisper
        self.model = whisper.load_model(model)
        self.language = language

    def transcribe(self, audio: np.array, **transcribe_options) -> str:
        result = self.model.transcribe(audio,
                                       without_timestamps=True,
                                       language=self.language,
                                       **transcribe_options)
        return result.get('text')

    def loop(self, input_queue: queue.SimpleQueue[TranslationTask],
             output_queue: queue.SimpleQueue[TranslationTask], whisper_filters: str,
             print_result: bool, output_timestamps: bool, **transcribe_options):
        while True:
            task = input_queue.get()
            task.transcribed_text = _filter_text(self.transcribe(task.audio, **transcribe_options),
                                                 whisper_filters).strip()
            if not task.transcribed_text:
                if print_result:
                    print('skip...')
                continue
            if print_result:
                if output_timestamps:
                    timestamp_text = '{} --> {}'.format(sec2str(task.time_range[0]),
                                                        sec2str(task.time_range[1]))
                    print(timestamp_text + ' ' + task.transcribed_text)
                else:
                    print(task.transcribed_text)
            output_queue.put(task)


class FasterWhisper(OpenaiWhisper):

    def __init__(self, model: str, language: str) -> None:
        print('Loading faster-whisper model: {}'.format(model))
        from faster_whisper import WhisperModel
        self.model = WhisperModel(model)
        self.language = language

    def transcribe(self, audio: np.array, **transcribe_options) -> str:
        segments, info = self.model.transcribe(audio, language=self.language, **transcribe_options)
        transcribed_text = ''
        for segment in segments:
            transcribed_text += segment.text
        return transcribed_text


class RemoteOpenaiWhisper(OpenaiWhisper):
    # https://platform.openai.com/docs/api-reference/audio/createTranscription?lang=python

    def __init__(self, language: str, proxy: str) -> None:
        self.client = OpenAI()
        self.client = OpenAI(http_client=DefaultHttpxClient(proxies=proxy))
        self.language = language

    def __del__(self):
        if os.path.exists(TEMP_AUDIO_FILE_NAME):
            os.remove(TEMP_AUDIO_FILE_NAME)

    def transcribe(self, audio: np.array, **transcribe_options) -> str:
        with open(TEMP_AUDIO_FILE_NAME, 'wb') as audio_file:
            write_audio(audio_file, SAMPLE_RATE, audio)
        with open(TEMP_AUDIO_FILE_NAME, 'rb') as audio_file:
            result = self.client.audio.transcriptions.create(model='whisper-1',
                                                             file=audio_file,
                                                             language=self.language).text
        os.remove(TEMP_AUDIO_FILE_NAME)
        return result



import json
import queue
import threading
import time
import re
from collections import deque
from datetime import datetime, timedelta

import google.generativeai as genai
from google.api_core.exceptions import InternalServerError, ResourceExhausted
from google.generativeai.types import HarmCategory, HarmBlockThreshold
from openai import OpenAI, DefaultHttpxClient, APITimeoutError, APIConnectionError

from .common import TranslationTask, LoopWorkerBase


# The double quotes in the values of JSON have not been escaped, so manual escaping is necessary.
def escape_specific_quotes(input_string):
    quote_positions = [i for i, char in enumerate(input_string) if char == '"']

    if len(quote_positions) <= 4:
        return input_string

    for i in range(3, len(quote_positions) - 1):
        position = quote_positions[i]
        input_string = input_string[:position] + '\\"' + input_string[position + 1:]
        quote_positions = [pos + 1 if pos > position else pos for pos in quote_positions]

    return input_string


def _parse_json_completion(completion):
    pattern = re.compile(r'\{.*}', re.DOTALL)
    json_match = pattern.search(completion)

    if not json_match:
        return completion

    json_str = json_match.group(0)
    json_str = escape_specific_quotes(json_str)

    try:
        json_obj = json.loads(json_str)
        translate_text = json_obj.get('translation', None)
        if not translate_text:
            return completion
        return translate_text
    except json.JSONDecodeError:
        return completion


class LLMClint():

    class LLM_TYPE:
        GPT = 'GPT'
        GEMINI = 'Gemini'

    def __init__(self, llm_type: str, model: str, prompt: str, history_size: int, proxy: str,
                 use_json_result: bool) -> None:
        if llm_type not in (self.LLM_TYPE.GPT, self.LLM_TYPE.GEMINI):
            raise ValueError('Unknow LLM type: {}'.format(llm_type))
        print('Using {} API as translation engine.'.format(model))
        self.llm_type = llm_type
        self.model = model
        self.prompt = prompt
        self.history_size = history_size
        self.history_messages = []
        self.proxy = proxy
        self.use_json_result = use_json_result

    def _append_history_message(self, user_content: str, assistant_content: str):
        if not user_content or not assistant_content:
            return
        self.history_messages.extend([{
            'role': 'user',
            'content': user_content
        }, {
            'role': 'assistant',
            'content': assistant_content
        }])
        while (len(self.history_messages) > self.history_size * 2):
            self.history_messages.pop(0)

    def _translate_by_gpt(self, translation_task: TranslationTask):
        # https://platform.openai.com/docs/api-reference/chat/create?lang=python
        client = OpenAI(http_client=DefaultHttpxClient(proxies=self.proxy))
        system_prompt = 'You are a translation engine.'
        if self.use_json_result:
            system_prompt += " Output the answer in json format, key is translation."
        messages = [{'role': 'system', 'content': system_prompt}]
        messages.extend(self.history_messages)
        user_content = '{}: \n{}'.format(self.prompt, translation_task.transcribed_text)
        messages.append({'role': 'user', 'content': user_content})
        try:
            completion = client.chat.completions.create(
                model=self.model,
                temperature=0,
                max_tokens=1000,
                top_p=1,
                frequency_penalty=1,
                presence_penalty=1,
                messages=messages,
            )

            translation_task.translated_text = completion.choices[0].message.content
            if self.use_json_result:
                translation_task.translated_text = _parse_json_completion(
                    translation_task.translated_text)
        except (APITimeoutError, APIConnectionError) as e:
            print(e)
            return
        if self.history_size:
            self._append_history_message(user_content, translation_task.translated_text)

    @staticmethod
    def _gpt_to_gemini(gpt_messages: list):
        gemini_messages = []
        for gpt_message in gpt_messages:
            gemini_message = {}
            gemini_message['role'] = gpt_message['role']
            if gemini_message['role'] == 'assistant':
                gemini_message['role'] = 'model'
            gemini_message['parts'] = [gpt_message['content']]
            gemini_messages.append(gemini_message)
        return gemini_messages

    def _translate_by_gemini(self, translation_task: TranslationTask):
        # https://ai.google.dev/tutorials/python_quickstart
        client = genai.GenerativeModel(self.model)
        messages = self._gpt_to_gemini(self.history_messages)
        user_content = '{}: \n{}'.format(self.prompt, translation_task.transcribed_text)
        messages.append({'role': 'user', 'parts': [user_content]})
        config = genai.types.GenerationConfig(candidate_count=1, temperature=0)
        safety_settings = {
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
        }
        try:
            response = client.generate_content(messages,
                                               generation_config=config,
                                               safety_settings=safety_settings)
            translation_task.translated_text = response.text
            if self.use_json_result:
                translation_task.translated_text = _parse_json_completion(
                    translation_task.translated_text)
        except (ValueError, InternalServerError, ResourceExhausted) as e:
            print(e)
            return
        if self.history_size:
            self._append_history_message(user_content, translation_task.translated_text)

    def translate(self, translation_task: TranslationTask):
        if self.llm_type == self.LLM_TYPE.GPT:
            self._translate_by_gpt(translation_task)
        elif self.llm_type == self.LLM_TYPE.GEMINI:
            self._translate_by_gemini(translation_task)
        else:
            raise ValueError('Unknow LLM type: {}'.format(self.llm_type))


class ParallelTranslator(LoopWorkerBase):
    PARALLEL_MAX_NUMBER = 10

    def __init__(self, llm_client: LLMClint, timeout: int, retry_if_translation_fails: bool):
        self.llm_client = llm_client
        self.timeout = timeout
        self.retry_if_translation_fails = retry_if_translation_fails
        self.processing_queue = deque()

    def trigger(self, translation_task: TranslationTask):
        self.processing_queue.append(translation_task)
        translation_task.start_time = datetime.utcnow()
        thread = threading.Thread(target=self.llm_client.translate, args=(translation_task,))
        thread.daemon = True
        thread.start()

    def get_results(self):
        results = []
        while self.processing_queue and (self.processing_queue[0].translated_text or
                                         datetime.utcnow() - self.processing_queue[0].start_time
                                         > timedelta(seconds=self.timeout)):
            task = self.processing_queue.popleft()
            if task.translated_text:
                results.append(task)
            else:
                if self.retry_if_translation_fails:
                    self.trigger(task)
                else:
                    results.append(task)
                    print('Translation timeout or failed: {}'.format(task.transcribed_text))
        return results

    def loop(self, input_queue: queue.SimpleQueue[TranslationTask],
             output_queue: queue.SimpleQueue[TranslationTask]):
        while True:
            if not input_queue.empty() and len(self.processing_queue) < self.PARALLEL_MAX_NUMBER:
                task = input_queue.get()
                self.trigger(task)
            finished_tasks = self.get_results()
            for task in finished_tasks:
                output_queue.put(task)
            time.sleep(0.1)


class SerialTranslator(LoopWorkerBase):

    def __init__(self, llm_client: LLMClint, timeout: int, retry_if_translation_fails: bool):
        self.llm_client = llm_client
        self.timeout = timeout
        self.retry_if_translation_fails = retry_if_translation_fails

    def trigger(self, translation_task: TranslationTask):
        translation_task.start_time = datetime.utcnow()
        thread = threading.Thread(target=self.llm_client.translate, args=(translation_task,))
        thread.daemon = True
        thread.start()

    def loop(self, input_queue: queue.SimpleQueue[TranslationTask],
             output_queue: queue.SimpleQueue[TranslationTask]):
        current_task = None
        while True:
            if current_task:
                if (current_task.translated_text or datetime.utcnow() - current_task.start_time
                        > timedelta(seconds=self.timeout)):
                    if not current_task.translated_text:
                        if self.retry_if_translation_fails:
                            self.trigger(current_task)
                            continue
                        print('Translation timeout or failed: {}'.format(
                            current_task.transcribed_text))
                    output_queue.put(current_task)
                    current_task = None

            if current_task is None and not input_queue.empty():
                current_task = input_queue.get()
                self.trigger(current_task)
            time.sleep(0.1)



import re


def emoji_filter(text: str):
    return re.sub(
        r'[\U0001F600-\U0001F64F\U0001F300-\U0001F5FF\U0001F680-\U0001F6FF\U0001F1E0-\U0001F1FF]+',
        '', text)


def japanese_stream_filter(text: str):
    for filter_pattern in [
            r'【.+】', r'ご視聴ありがとうございました', r'チャンネル登録をお願いいたします', r'ご視聴いただきありがとうございます', r'チャンネル登録してね',
            r'字幕視聴ありがとうございました', r'動画をご覧頂きましてありがとうございました', r'次の動画でお会いしましょう', r'最後までご視聴頂きありがとうございました',
            r'次の動画もお楽しみに', r'次回もお楽しみに', r'また次回の動画でお会いしましょう', r'ご覧いただきありがとうございます',
            r'最後までご視聴頂き有難うございました', r'最後までご視聴頂き有難う御座いました', r'チャンネル登録よろしくお願いします'
    ]:
        text = re.sub(filter_pattern, '', text)

    for filter_text in ['字幕作成', 'この動画の字幕', 'by ']:
        if filter_text in text:
            print('filter', text)
            return ''

    for filter_text in ['エンディング', '次回予告', 'またね', 'ありがとうございました', 'それではまた']:
        if filter_text == text:
            print('filter', text)
            return ''

    if len(text) < 3:
        return ''
    return text



import os
import queue
import signal
import subprocess
import sys
import threading

import ffmpeg
import numpy as np

from .common import SAMPLE_RATE, LoopWorkerBase


def _transport(ytdlp_proc, ffmpeg_proc):
    while (ytdlp_proc.poll() is None) and (ffmpeg_proc.poll() is None):
        try:
            chunk = ytdlp_proc.stdout.read(1024)
            ffmpeg_proc.stdin.write(chunk)
        except (BrokenPipeError, OSError):
            pass
    ytdlp_proc.kill()
    ffmpeg_proc.kill()


def _open_stream(url: str, format: str, cookies: str, proxy: str):
    cmd = ['yt-dlp', url, '-f', format, '-o', '-', '-q']
    if cookies:
        cmd.extend(['--cookies', cookies])
    if proxy:
        cmd.extend(['--proxy', proxy])
    ytdlp_process = subprocess.Popen(cmd, stdout=subprocess.PIPE)

    try:
        ffmpeg_process = (ffmpeg.input('pipe:', loglevel='panic').output('pipe:',
                                                                         format='s16le',
                                                                         acodec='pcm_s16le',
                                                                         ac=1,
                                                                         ar=SAMPLE_RATE).run_async(
                                                                             pipe_stdin=True,
                                                                             pipe_stdout=True))
    except ffmpeg.Error as e:
        raise RuntimeError(f'Failed to load audio: {e.stderr.decode()}') from e

    thread = threading.Thread(target=_transport, args=(ytdlp_process, ffmpeg_process))
    thread.start()
    return ffmpeg_process, ytdlp_process


class StreamAudioGetter(LoopWorkerBase):

    def __init__(self, url: str, format: str, cookies: str, proxy: str,
                 frame_duration: float) -> None:
        self._cleanup_ytdlp_cache()

        print('Opening stream: {}'.format(url))
        self.ffmpeg_process, self.ytdlp_process = _open_stream(url, format, cookies, proxy)
        self.byte_size = round(frame_duration * SAMPLE_RATE *
                               2)  # Factor 2 comes from reading the int16 stream as bytes
        signal.signal(signal.SIGINT, self._exit_handler)

    def __del__(self):
        self._cleanup_ytdlp_cache()

    def _exit_handler(self, signum, frame):
        self.ffmpeg_process.kill()
        if self.ytdlp_process:
            self.ytdlp_process.kill()
        sys.exit(0)

    def _cleanup_ytdlp_cache(self):
        for file in os.listdir('./'):
            if file.startswith('--Frag'):
                os.remove(file)

    def loop(self, output_queue: queue.SimpleQueue[np.array]):
        while self.ffmpeg_process.poll() is None:
            in_bytes = self.ffmpeg_process.stdout.read(self.byte_size)
            if not in_bytes:
                break
            if len(in_bytes) != self.byte_size:
                continue
            audio = np.frombuffer(in_bytes, np.int16).flatten().astype(np.float32) / 32768.0
            output_queue.put(audio)

        self.ffmpeg_process.kill()
        if self.ytdlp_process:
            self.ytdlp_process.kill()


class LocalFileAudioGetter(LoopWorkerBase):

    def __init__(self, file_path: str, frame_duration: float) -> None:
        print('Opening local file: {}'.format(file_path))
        try:
            self.ffmpeg_process = (ffmpeg.input(
                file_path, loglevel='panic').output('pipe:',
                                                    format='s16le',
                                                    acodec='pcm_s16le',
                                                    ac=1,
                                                    ar=SAMPLE_RATE).run_async(pipe_stdin=True,
                                                                              pipe_stdout=True))
        except ffmpeg.Error as e:
            raise RuntimeError(f'Failed to load audio: {e.stderr.decode()}') from e
        self.byte_size = round(frame_duration * SAMPLE_RATE *
                               2)  # Factor 2 comes from reading the int16 stream as bytes
        signal.signal(signal.SIGINT, self._exit_handler)

    def _exit_handler(self, signum, frame):
        self.ffmpeg_process.kill()
        sys.exit(0)

    def loop(self, output_queue: queue.SimpleQueue[np.array]):
        while self.ffmpeg_process.poll() is None:
            in_bytes = self.ffmpeg_process.stdout.read(self.byte_size)
            if not in_bytes:
                break
            if len(in_bytes) != self.byte_size:
                continue
            audio = np.frombuffer(in_bytes, np.int16).flatten().astype(np.float32) / 32768.0
            output_queue.put(audio)

        self.ffmpeg_process.kill()


class DeviceAudioGetter(LoopWorkerBase):

    def __init__(self, device_index: int, frame_duration: float) -> None:
        import sounddevice as sd
        if device_index:
            sd.default.device[0] = device_index
        sd.default.dtype[0] = np.float32
        self.frame_duration = frame_duration
        print('Recording device: {}'.format(sd.query_devices(sd.default.device[0])['name']))

    def loop(self, output_queue: queue.SimpleQueue[np.array]):
        import sounddevice as sd
        while True:
            audio = sd.rec(frames=round(SAMPLE_RATE * self.frame_duration),
                           samplerate=SAMPLE_RATE,
                           channels=1,
                           blocking=True).flatten()
            output_queue.put(audio)



import argparse
import os
import queue
import sys
import threading
import time

import google.generativeai as genai
from google.api_core.client_options import ClientOptions

from .audio_getter import StreamAudioGetter, LocalFileAudioGetter, DeviceAudioGetter
from .audio_slicer import AudioSlicer
from .audio_transcriber import OpenaiWhisper, FasterWhisper, RemoteOpenaiWhisper
from .llm_translator import LLMClint, ParallelTranslator, SerialTranslator
from .result_exporter import ResultExporter


def _start_daemon_thread(func, *args, **kwargs):
    thread = threading.Thread(target=func, args=args, kwargs=kwargs)
    thread.daemon = True
    thread.start()


def main(url, format, cookies, input_proxy, device_index, frame_duration,
         continuous_no_speech_threshold, min_audio_length, max_audio_length,
         prefix_retention_length, vad_threshold, model, language, use_faster_whisper,
         use_whisper_api, whisper_filters, openai_api_key, google_api_key, gpt_translation_prompt,
         gpt_translation_history_size, gpt_model, gemini_model, gpt_translation_timeout,
         gpt_base_url, gemini_base_url, processing_proxy, use_json_result,
         retry_if_translation_fails, output_timestamps, hide_transcribe_result, output_proxy,
         output_file_path, cqhttp_url, cqhttp_token, discord_webhook_url, telegram_token,
         telegram_chat_id, **transcribe_options):
    if openai_api_key:
        os.environ['OPENAI_API_KEY'] = openai_api_key
    if gpt_base_url:
        os.environ['OPENAI_BASE_URL'] = gpt_base_url
    if google_api_key:
        gemini_client_options = ClientOptions(api_endpoint=gemini_base_url)
        genai.configure(api_key=google_api_key,
                        client_options=gemini_client_options,
                        transport='rest')

    getter_to_slicer_queue = queue.SimpleQueue()
    slicer_to_transcriber_queue = queue.SimpleQueue()
    transcriber_to_translator_queue = queue.SimpleQueue()
    translator_to_exporter_queue = queue.SimpleQueue(
    ) if gpt_translation_prompt else transcriber_to_translator_queue

    _start_daemon_thread(
        ResultExporter.work,
        output_whisper_result=not hide_transcribe_result,
        output_timestamps=output_timestamps,
        proxy=output_proxy,
        output_file_path=output_file_path,
        cqhttp_url=cqhttp_url,
        cqhttp_token=cqhttp_token,
        discord_webhook_url=discord_webhook_url,
        telegram_token=telegram_token,
        telegram_chat_id=telegram_chat_id,
        input_queue=translator_to_exporter_queue,
    )
    if gpt_translation_prompt:
        if google_api_key:
            llm_client = LLMClint(
                llm_type=LLMClint.LLM_TYPE.GEMINI,
                model=gemini_model,
                prompt=gpt_translation_prompt,
                history_size=gpt_translation_history_size,
                proxy=processing_proxy,
                use_json_result=use_json_result,
            )
        else:
            llm_client = LLMClint(
                llm_type=LLMClint.LLM_TYPE.GPT,
                model=gpt_model,
                prompt=gpt_translation_prompt,
                history_size=gpt_translation_history_size,
                proxy=processing_proxy,
                use_json_result=use_json_result,
            )
        if gpt_translation_history_size == 0:
            _start_daemon_thread(
                ParallelTranslator.work,
                llm_client=llm_client,
                timeout=gpt_translation_timeout,
                retry_if_translation_fails=retry_if_translation_fails,
                input_queue=transcriber_to_translator_queue,
                output_queue=translator_to_exporter_queue,
            )
        else:
            _start_daemon_thread(
                SerialTranslator.work,
                llm_client=llm_client,
                timeout=gpt_translation_timeout,
                retry_if_translation_fails=retry_if_translation_fails,
                input_queue=transcriber_to_translator_queue,
                output_queue=translator_to_exporter_queue,
            )
    if use_faster_whisper:
        _start_daemon_thread(FasterWhisper.work,
                             model=model,
                             language=language,
                             print_result=not hide_transcribe_result,
                             output_timestamps=output_timestamps,
                             input_queue=slicer_to_transcriber_queue,
                             output_queue=transcriber_to_translator_queue,
                             whisper_filters=whisper_filters,
                             **transcribe_options)
    elif use_whisper_api:
        _start_daemon_thread(RemoteOpenaiWhisper.work,
                             language=language,
                             proxy=processing_proxy,
                             print_result=not hide_transcribe_result,
                             output_timestamps=output_timestamps,
                             input_queue=slicer_to_transcriber_queue,
                             output_queue=transcriber_to_translator_queue,
                             whisper_filters=whisper_filters,
                             **transcribe_options)
    else:
        _start_daemon_thread(OpenaiWhisper.work,
                             model=model,
                             language=language,
                             print_result=not hide_transcribe_result,
                             output_timestamps=output_timestamps,
                             input_queue=slicer_to_transcriber_queue,
                             output_queue=transcriber_to_translator_queue,
                             whisper_filters=whisper_filters,
                             **transcribe_options)
    _start_daemon_thread(
        AudioSlicer.work,
        frame_duration=frame_duration,
        continuous_no_speech_threshold=continuous_no_speech_threshold,
        min_audio_length=min_audio_length,
        max_audio_length=max_audio_length,
        prefix_retention_length=prefix_retention_length,
        vad_threshold=vad_threshold,
        input_queue=getter_to_slicer_queue,
        output_queue=slicer_to_transcriber_queue,
    )
    if url.lower() == 'device':
        DeviceAudioGetter.work(
            device_index=device_index,
            frame_duration=frame_duration,
            output_queue=getter_to_slicer_queue,
        )
    elif os.path.isabs(url):
        LocalFileAudioGetter.work(
            file_path=url,
            frame_duration=frame_duration,
            output_queue=getter_to_slicer_queue,
        )
    else:
        StreamAudioGetter.work(
            url=url,
            format=format,
            cookies=cookies,
            proxy=input_proxy,
            frame_duration=frame_duration,
            output_queue=getter_to_slicer_queue,
        )

    # Wait for others process finish.
    while (not getter_to_slicer_queue.empty() or not slicer_to_transcriber_queue.empty() or
           not transcriber_to_translator_queue.empty() or not translator_to_exporter_queue.empty()):
        time.sleep(5)
    print('Stream ended')


def cli():
    parser = argparse.ArgumentParser(description='Parameters for translator.py')
    parser.add_argument('URL',
                        type=str,
                        help='The URL of the stream. '
                        'If a local file path is filled in, it will be used as input. '
                        'If fill in "device", the input will be obtained from your PC device.')
    parser.add_argument('--format',
                        type=str,
                        default='wa*',
                        help='Stream format code, '
                        'this parameter will be passed directly to yt-dlp.')
    parser.add_argument('--cookies',
                        type=str,
                        default=None,
                        help='Used to open member-only stream, '
                        'this parameter will be passed directly to yt-dlp.')
    parser.add_argument('--input_proxy',
                        type=str,
                        default=None,
                        help='Use the specified HTTP/HTTPS/SOCKS proxy for yt-dlp, '
                        'e.g. http://127.0.0.1:7890.')
    parser.add_argument('--device_index',
                        type=int,
                        default=None,
                        help='The index of the device that needs to be recorded. '
                        'If not set, the system default recording device will be used.')
    parser.add_argument('--print_all_devices',
                        action='store_true',
                        help='Print all audio devices info then exit.')
    parser.add_argument('--frame_duration',
                        type=float,
                        default=0.1,
                        help='The unit that processes live streaming data in seconds, '
                        'should be >= 0.03')
    parser.add_argument('--continuous_no_speech_threshold',
                        type=float,
                        default=0.5,
                        help='Slice if there is no speech for a continuous period in second.')
    parser.add_argument('--min_audio_length',
                        type=float,
                        default=1.5,
                        help='Minimum slice audio length in seconds.')
    parser.add_argument('--max_audio_length',
                        type=float,
                        default=15.0,
                        help='Maximum slice audio length in seconds.')
    parser.add_argument('--prefix_retention_length',
                        type=float,
                        default=0.5,
                        help='The length of the retention prefix audio during slicing.')
    parser.add_argument('--vad_threshold',
                        type=float,
                        default=0.35,
                        help='The threshold of Voice activity detection.'
                        'if the speech probability of a frame is higher than this value, '
                        'then this frame is speech.')
    parser.add_argument(
        '--model',
        type=str,
        default='small',
        help='Select Whisper/Faster-Whisper model size. '
        'See https://github.com/openai/whisper#available-models-and-languages for available models.'
    )
    parser.add_argument(
        '--language',
        type=str,
        default='auto',
        help='Language spoken in the stream. '
        'Default option is to auto detect the spoken language. '
        'See https://github.com/openai/whisper#available-models-and-languages for available languages.'
    )
    parser.add_argument('--beam_size',
                        type=int,
                        default=5,
                        help='Number of beams in beam search. '
                        'Set to 0 to use greedy algorithm instead.')
    parser.add_argument('--best_of',
                        type=int,
                        default=5,
                        help='Number of candidates when sampling with non-zero temperature.')
    parser.add_argument('--use_faster_whisper',
                        action='store_true',
                        help='Set this flag to use faster-whisper implementation instead of '
                        'the original OpenAI implementation.')
    parser.add_argument('--use_whisper_api',
                        action='store_true',
                        help='Set this flag to use OpenAI Whisper API instead of '
                        'the original local Whipser.')
    parser.add_argument('--whisper_filters',
                        type=str,
                        default='emoji_filter',
                        help='Filters apply to whisper results, separated by ",". '
                        'We provide emoji_filter and japanese_stream_filter.')
    parser.add_argument('--openai_api_key',
                        type=str,
                        default=None,
                        help='OpenAI API key if using GPT translation / Whisper API.')
    parser.add_argument('--google_api_key',
                        type=str,
                        default=None,
                        help='Google API key if using Gemini translation.')
    parser.add_argument('--gpt_model',
                        type=str,
                        default='gpt-4o-mini',
                        help='OpenAI\'s GPT model name, gpt-3.5-turbo / gpt-4o / gpt-4o-mini.')
    parser.add_argument('--gemini_model',
                        type=str,
                        default='gemini-1.5-flash',
                        help='Google\'s Gemini model name, gemini-1.5-flash / gemini-1.5-pro')
    parser.add_argument(
        '--gpt_translation_prompt',
        type=str,
        default=None,
        help='If set, will translate result text to target language via GPT / Gemini API. '
        'Example: \"Translate from Japanese to Chinese\"')
    parser.add_argument(
        '--gpt_translation_history_size',
        type=int,
        default=0,
        help='The number of previous messages sent when calling the GPT / Gemini API. '
        'If the history size is 0, the translation will be run parallelly. '
        'If the history size > 0, the translation will be run serially.')
    parser.add_argument('--gpt_translation_timeout',
                        type=int,
                        default=10,
                        help='If the GPT / Gemini translation exceeds this number of seconds, '
                        'the translation will be discarded.')
    parser.add_argument('--gpt_base_url',
                        type=str,
                        default=None,
                        help='Customize the API endpoint of GPT.')
    parser.add_argument('--gemini_base_url',
                        type=str,
                        default=None,
                        help='Customize the API endpoint of Gemini.')
    parser.add_argument(
        '--processing_proxy',
        type=str,
        default=None,
        help='Use the specified HTTP/HTTPS/SOCKS proxy for Whisper/GPT API '
        '(Gemini currently doesn\'t support specifying a proxy within the program), '
        'e.g. http://127.0.0.1:7890.')
    parser.add_argument(
        '--use_json_result',
        action='store_true',
        help='Using JSON result in LLM translation for some locally deployed models.')
    parser.add_argument(
        '--retry_if_translation_fails',
        action='store_true',
        help='Retry when translation times out/fails. Used to generate subtitles offline.')
    parser.add_argument('--output_timestamps',
                        action='store_true',
                        help='Output the timestamp of the text when outputting the text.')
    parser.add_argument('--hide_transcribe_result',
                        action='store_true',
                        help='Hide the result of Whisper transcribe.')
    parser.add_argument(
        '--output_proxy',
        type=str,
        default=None,
        help='Use the specified HTTP/HTTPS/SOCKS proxy for Cqhttp/Discord/Telegram, '
        'e.g. http://127.0.0.1:7890.')
    parser.add_argument('--output_file_path',
                        type=str,
                        default=None,
                        help='If set, will save the result text to this path.')
    parser.add_argument('--cqhttp_url',
                        type=str,
                        default=None,
                        help='If set, will send the result text to this Cqhttp server.')
    parser.add_argument('--cqhttp_token',
                        type=str,
                        default=None,
                        help='Token of cqhttp, if it is not set on the server side, '
                        'it does not need to fill in.')
    parser.add_argument('--discord_webhook_url',
                        type=str,
                        default=None,
                        help='If set, will send the result text to this Discord channel.')
    parser.add_argument('--telegram_token', type=str, default=None, help='Token of Telegram bot.')
    parser.add_argument('--telegram_chat_id',
                        type=int,
                        default=None,
                        help='If set, will send the result text to this Telegram chat. '
                        'Needs to be used with \"--telegram_token\".')

    args = parser.parse_args().__dict__
    url = args.pop('URL')

    if args['print_all_devices']:
        import sounddevice as sd
        print(sd.query_devices())
        exit(0)

    if args['model'].endswith('.en'):
        if args['model'] == 'large.en':
            print(
                'English model does not have large model, please choose from {tiny.en, small.en, medium.en}'
            )
            sys.exit(0)
        if args['language'] != 'English' and args['language'] != 'en':
            if args['language'] == 'auto':
                print('Using .en model, setting language from auto to English')
                args['language'] = 'en'
            else:
                print(
                    'English model cannot be used to detect non english language, please choose a non .en model'
                )
                sys.exit(0)

    if args['use_faster_whisper'] and args['use_whisper_api']:
        print('Cannot use Faster Whisper and Whisper API at the same time')
        sys.exit(0)

    if args['use_whisper_api'] and not args['openai_api_key']:
        print('Please fill in the OpenAI API key when enabling Whisper API')
        sys.exit(0)

    if args['gpt_translation_prompt'] and not (args['openai_api_key'] or args['google_api_key']):
        print('Please fill in the OpenAI / Google API key when enabling LLM translation')
        sys.exit(0)

    if args['language'] == 'auto':
        args['language'] = None

    if args['beam_size'] == 0:
        args['beam_size'] = None

    main(url, **args)



import os
import queue
import requests

from .common import TranslationTask, LoopWorkerBase, sec2str


def _send_to_cqhttp(url: str, token: str, proxies: dict, text: str):
    headers = {'Authorization': 'Bearer {}'.format(token)} if token else None
    data = {'message': text}
    try:
        requests.post(url, headers=headers, data=data, timeout=10, proxies=proxies)
    except Exception as e:
        print(e)


def _send_to_discord(webhook_url: str, proxies: dict, text: str):
    for sub_text in text.split('\n'):
        data = {'content': sub_text}
        try:
            requests.post(webhook_url, json=data, timeout=10, proxies=proxies)
        except Exception as e:
            print(e)


def _send_to_telegram(token: str, chat_id: int, proxies: dict, text: str):
    url = 'https://api.telegram.org/bot{}/sendMessage?chat_id={}&text={}'.format(
        token, chat_id, text)
    try:
        requests.post(url, timeout=10, proxies=proxies)
    except Exception as e:
        print(e)


def _output_to_file(file_path: str, text: str):
    with open(file_path, 'a', encoding='utf-8') as f:
        f.write(text + '\n\n')


class ResultExporter(LoopWorkerBase):

    def __init__(self, output_file_path: str) -> None:
        if output_file_path:
            if os.path.exists(output_file_path):
                os.remove(output_file_path)

    def loop(self, input_queue: queue.SimpleQueue[TranslationTask], output_whisper_result: bool,
             output_timestamps: bool, proxy: str, output_file_path: str, cqhttp_url: str,
             cqhttp_token: str, discord_webhook_url: str, telegram_token: str,
             telegram_chat_id: int):
        proxies = {"http": proxy, "https": proxy} if proxy else None
        while True:
            task = input_queue.get()
            timestamp_text = '{} --> {}'.format(sec2str(task.time_range[0]),
                                                sec2str(task.time_range[1]))
            text_to_send = (task.transcribed_text + '\n') if output_whisper_result else ''
            if output_timestamps:
                text_to_send = timestamp_text + '\n' + text_to_send
            if task.translated_text:
                text_to_print = task.translated_text
                if output_timestamps:
                    text_to_print = timestamp_text + ' ' + text_to_print
                print('\033[1m{}\033[0m'.format(text_to_print))
                text_to_send += task.translated_text
            text_to_send = text_to_send.strip()
            if output_file_path:
                _output_to_file(output_file_path, text_to_send)
            if cqhttp_url:
                _send_to_cqhttp(cqhttp_url, cqhttp_token, proxies, text_to_send)
            if discord_webhook_url:
                _send_to_discord(discord_webhook_url, proxies, text_to_send)
            if telegram_token and telegram_chat_id:
                _send_to_telegram(telegram_token, telegram_chat_id, proxies, text_to_send)



from abc import ABC, abstractmethod
from datetime import datetime

import numpy as np
from whisper.audio import SAMPLE_RATE


class TranslationTask:

    def __init__(self, audio: np.array, time_range: tuple[float, float]):
        self.audio = audio
        self.transcribed_text = None
        self.translated_text = None
        self.time_range = time_range
        self.start_time = None


def _auto_args(func, kwargs):
    names = func.__code__.co_varnames
    return {k: v for k, v in kwargs.items() if k in names}


class LoopWorkerBase(ABC):

    @abstractmethod
    def loop(self):
        pass

    @classmethod
    def work(cls, **kwargs):
        obj = cls(**_auto_args(cls.__init__, kwargs))
        obj.loop(**_auto_args(obj.loop, kwargs))


def sec2str(second: float):
    dt = datetime.utcfromtimestamp(second)
    result = dt.strftime('%H:%M:%S')
    result += ',' + str(int(second * 10 % 10))
    return result
